"""
Documents' module unit tests.
"""
import unittest
from os import path

from netcdfella.document import Document, NetCDF


class TestDocument(unittest.TestCase):
    """
    The test for class Document.
    """

    def test_constructor(self):
        "test_constructor checks if the Document gets initialized correctly."
        test_doc = Document("test", "/")
        self.assertEqual(test_doc.name, "test")
        self.assertEqual(test_doc.path, "/")
        self.assertTrue(isinstance(test_doc, Document))

    def test_read(self):
        "test_read checks if a netcdf file can be read."
        doc_path = path.join(path.dirname(__file__), "test_netcdf4.nc")
        test_doc = NetCDF("test", doc_path)
        test_doc.read()

    def test_get_dim(self):
        "test_get_dim returns the dimension of a netcdf file."
        doc_path = path.join(path.dirname(__file__), "test_netcdf4.nc")
        test_doc = NetCDF("test", doc_path)
        test_doc.read()
        test_doc.get_dim("flashes")

    def test__get_variable_samples(self):
        "test_get_variable_samples tests the return of sample values."
        doc_path = path.join(path.dirname(__file__), "test_netcdf4.nc")
        test_doc = NetCDF("test", doc_path)
        test_doc.read()
        test_doc._get_variable_samples("flashes", "radiance", 5)

    def test_to_ascii(self):
        "test_to_ascii writes a netcdf as ascii."
        doc_path = path.join(path.dirname(__file__), "test_netcdf4.nc")
        test_doc = NetCDF("test", doc_path)
        test_doc.read()
        test_doc.to_ascii()

    def test_to_img_scatter(self):
        """
        This function takes a list of x and y coordinates
        and a list of colors and plots them on a scatter
        plot.
        """
        "test_to_ascii writes a netcdf as ascii."
        doc_path = path.join(path.dirname(__file__), "test_netcdf4.nc")
        test_doc = NetCDF("test", doc_path)
        test_doc.read()
        test_doc.to_img_scatter("title", "flashes",
                                "longitude", "latitude", "radiance")

    def test_to_img_marks(self):
        """
        This function takes a list of x and y coordinates scatters
        marks on the map plot.
        """
        "test_to_ascii writes a netcdf as ascii."
        doc_path = path.join(path.dirname(__file__), "test_netcdf4.nc")
        test_doc = NetCDF("test", doc_path)
        test_doc.read()
        test_doc.to_img_marks("title", "longitude", "latitude", "radiance")

    def test_to_graph(self):
        "test_to_graph checks the function to_graph."
        doc_path = path.join(path.dirname(__file__), "test_netcdf4.nc")
        test_doc = NetCDF("test", doc_path)
        test_doc.read()
        test_doc.to_graph()



"""
Document module defines documents that get manipulated by netcdfella.
"""
from functools import partial
from os import path

import matplotlib.pyplot as plt
import numpy as np
from ncplot import view
from netCDF4 import Dataset

from netcdfella.maps import GeoStationaryMap


class Document:
    """
    Document is the generic form of netcdfella documents.
    """

    def __init__(self, name, path):
        self.name = name
        self.path = path
        self.input_kind = None
        self.output_path = None

    def set_output_kind(self, output_kind: list):
        "Set the format type for the converted document"
        self.output_kind = output_kind

    def set_input_kind(self, input_kind: str):
        "Set the format type for the input document"
        self.input_kind = input_kind

    def set_output_path(self, output_path):
        "Set the output path for converted files."
        self.output_path = output_path


class NetCDF(Document):
    "NetCDF class contains usefull info for NetCDF files."

    def __init__(self, name, doc_path):
        Document.__init__(self, name, doc_path)
        self.excluded_variables = {"timeliness_non_nominal"}
        self.dimensions = []
        self.dataset = None

    def exclude_variables(self, exclusions):
        "exclude netcdf variables from netcdf file reading."
        self.excluded_variables.update(exclusions)

    def exclude_variables_from_str(self, exclusions):
        """
        exclude netcdf variables from netcdf file reading
        from a comma separated list as string.
        """
        exc = exclusions.strip().split(",")
        self.excluded_variables.update(exc)

    def get_dim(self, name):
        "get_dim returns a dimension by name"
        for dim in self.dimensions:
            if dim.name == name:
                return dim

    def read(self):
        "Reads the contents of a netcdf file."
        self.dataset = Dataset(self.path)
        # pylint: disable=not-an-iterable
        for dim in self.dataset.dimensions:
            # pylint: disable=unsubscriptable-object
            new_dimension = self.Dimension(
                self.dataset.dimensions[dim].name,
                self.dataset.dimensions[dim].size
            )
            # pylint: disable=no-member
            for k in self.dataset.variables.keys():
                if k not in self.excluded_variables:
                    # pylint: disable=unsubscriptable-object
                    new_dimension.add_variables(k,
                                                self.dataset.variables[k][:])
            self.dimensions.append(new_dimension)

    def to_ascii(self):
        "to_ascii converts a netcdf document to ascii."
        open_utf8 = partial(open, encoding="UTF-8")
        ascii_file = path.splitext(self.path)[0] + ".asc"
        ascii_file = open_utf8(ascii_file, "w")
        first_line = ""
        for _, dim in enumerate(self.dimensions):
            first_line = "@DIMENSION: " + dim.name + " "
            ascii_file.write(first_line)
            ascii_file.write("\n")
            second_line = ""
            for key, _ in dim.variables.items():
                second_line = second_line + key + " "
            ascii_file.write("@VARIABLES: " + second_line)
            ascii_file.write("\n")
            ascii_file.write("@DATA\n")
            for i in range(dim.size):
                current_line = ""
                for _, var in dim.variables.items():
                    current_line = current_line + f"{var[i]}" + " "
                ascii_file.write(current_line)
                ascii_file.write("\n")
        ascii_file.close()

    def to_graph(self, variable=None):
        "to_graph creates graphical representations of variables."
        if self.output_path is not None:
            out_dir = self.output_path + path.splitext(self.path)[0] + ".html"
        else:
            out_dir = path.splitext(self.path)[0] + ".html"
        view(self.path, var=variable, out=out_dir, quadmesh=True)

    def to_img_scatter(self, title, dim_name, longitude_name,
                       latitude_name, variable_name, resolution="i",
                       latitude0=0, img_ext="png", plot_samples=3,
                       height=35786000, sphere=(6378137, 6356752.3142),
                       dot_scale=0.01, dot_transparency=0.2, color="magenta",
                       output_path="./", name_suffix="_scatter"):
        "to_jpeg converts a netcdf image to jpeg"
        plt.clf()
        output_name = output_path \
            + title.removesuffix('.nc') \
            + name_suffix+"."+img_ext
        map = self._create_map("geos", title+name_suffix, resolution,
                               latitude0, height, sphere)
        lons, lats = map(self.get_dim(dim_name).get_variable(longitude_name),
                         self.get_dim(dim_name).get_variable(latitude_name))
        map.scatter(lons, lats, latlon=False,
                    s=dot_scale*self.get_dim(dim_name)
                                    .get_variable(variable_name),
                    c=color,
                    alpha=dot_transparency)
        plot_samples = self._get_variable_samples(dim_name,
                                                  variable_name,
                                                  plot_samples)
        for a in plot_samples:
            map.scatter([], [], c=color, alpha=1, s=dot_scale * a,
                        label=str(a) + " "+variable_name)
        plt.legend(scatterpoints=1, frameon=True,
                   markerscale=2, labelspacing=1, loc=3)
        plt.title(title+name_suffix)
        plt.savefig(output_name)

    def to_img_marks(self, title, dim_name, longitude_name, latitude_name,
                     variable_name, img_ext="png", resolution="i",
                     latitude0=0, height=35786000, marker="x", color="b",
                     sphere=(6378137, 6356752.3142), output_path="./",
                     name_suffix="_marks"):
        """
        Creates an image with marks on the map.
        """
        plt.clf()
        output_name = output_path \
            + title.removesuffix('.nc') \
            + name_suffix+"."+img_ext
        map = self._create_map("geos", title+name_suffix, resolution,
                               latitude0, height, sphere)
        lons, lats = map(self.get_dim(dim_name).get_variable(longitude_name),
                         self.get_dim(dim_name).get_variable(latitude_name))
        map.scatter(lons, lats, marker=marker, color=color)
        plt.title(title+name_suffix)
        plt.savefig(output_name)

    def _create_map(self, kind, title, resolution, latitude0, height, sphere):
        "Creates the instance of a basemap."
        map = None
        if kind == "geos":
            map = GeoStationaryMap(title, resolution, latitude0)
        map.set_sat_height(height).set_sphere(sphere).create_map()
        map.draw()
        return map

    def _get_variable_samples(self, dimension, variable, samples):
        """
        _get_variable_samples returns some samples of the variable from
        smaller to bigger value.
        """
        var = self.get_dim(dimension).get_variable(variable)
        var.sort()
        split_ar = np.split(np.array(var), samples)
        sample_vals = []
        for ar in split_ar:
            sample_vals.append(ar[len(ar)//2])
        return sample_vals

    class Dimension:
        "Dimension describes each dimension of a netcdf file."

        def __init__(self, name, size):
            self.name = name
            self.size = size
            self.variables = {}

        def add_variables(self, name, variable):
            "adds variable data from netcdf file as dictionary."
            self.variables[name] = variable

        def get_variable(self, name):
            "returns the variable of a dimension"
            return self.variables[name]



"""
conversion module provides the tools around converions.
"""
from os import path

from netcdfella.document import NetCDF


class Conversion:
    """
    Conversion is the class about the converion process.
    """

    def __init__(self, output_path="./", input_format="nc"):
        self.input_kind = input_format
        self.document_template = None
        self.output_path = output_path
        self._create_ascii = False
        self._create_scatter_map = False
        self._create_graph = False
        self._create_marks_map = False
        self.map_dimension = None
        self.map_variable = None
        self.map_longitude = None
        self.map_latitude = None
        self.__init_document_template()

    def __init_document_template(self):
        "initializes the document kind template."
        if self.input_kind == "nc":
            self.document_template = NetCDF("template", "temlate")

    def enable_scatter_map(self):
        """
        enables the conversion to scatter map.
        """
        self._create_scatter_map = True

    def enable_marks_map(self):
        """
        enables the conversion to marks map.
        """
        self._create_marks_map = True

    def set_map_vars(self, dimension=None, variable=None,
                     longitude="longitude", latitude="latitude"):
        "sets the variables for a map."
        self.map_dimension = dimension
        self.map_variable = variable
        self.map_longitude = longitude
        self.map_latitude = latitude

    def enable_ascii(self):
        "enables the conversion to ascii."
        self._create_ascii = True

    def enable_graph(self):
        "enables the conversion to graph."
        self._create_graph = True

    def convert_document(self, document):
        "converts a document to the preselected format."
        _, file_extension = path.splitext(document)
        if file_extension.replace(".", "") != self.input_kind:
            return
        if self.input_kind == "nc":
            document = NetCDF(path.basename(document), document)
            document.read()
            if self._create_ascii:
                print(f"    *** [ converting {document.name} to ascii ]")
                document.to_ascii()
            if self._create_graph:
                print(f"    *** [ converting {document.name} to graph ]")
                document.to_graph()
            if self._create_scatter_map:
                print(f"    *** [ converting {document.name} to scatter map ]")
                document.to_img_scatter(document.name, self.map_dimension,
                                        self.map_longitude, self.map_latitude,
                                        self.map_variable)
            if self._create_marks_map:
                print(f"    *** [ converting {document.name} to marks map ]")
                document.to_img_marks(document.name, self.map_dimension,
                                      self.map_longitude, self.map_latitude,
                                      self.map_variable)

    def enable_output_from_str(self, output_kinds):
        """
        Enable kinds from a comma separated string of output kinds.
        For exampe 'ascii,graph'.
        """
        kinds = output_kinds.strip().split(",")
        for k in kinds:
            if k == "ascii":
                self.enable_ascii()
            if k == "graph":
                self.enable_graph()
            if k == "scatter":
                self.enable_scatter_map()
            if k == "marks":
                self.enable_marks_map()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['netcdfella']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['netcdfella = cli.__main__:main']}

setup_kwargs = {
    'name': 'netcdfella',
    'version': '0.2.1',
    'description': 'Netcdfella is providing multiple ways to convert netcdf filed into other data types such as ASCII, PNG and JPG, and even create graphs.',
    'long_description': '<a name="readme-top"></a>\n\n[![Issues][issues-shield]][issues-url]\n[![GPL License][license-shield]][license-url]\n[![LinkedIn][linkedin-shield]][linkedin-url]\n\n<!-- PROJECT LOGO -->\n<br />\n<div align="center">\n  <a href="https://github.com/nikfot/netcdfella">\n    <img src="assets/netcdfella_logo.png" alt="Logo" height="300" width="auto">\n  </a>\n\n  <h3 align="center">NetcdFella</h3>\n\n  <p align="center">\n    The easy as duck duck netcdf converter to ASCII, JPEG and more...\n    <br />\n    <a href="https://github.com/nikfot/netcdfella"><strong>Explore the docs »</strong></a>\n    <br />\n    <br />\n    ·\n    <a href="https://github.com/nikfot/netcdfella/issues">Report Bug</a>\n    ·\n    <a href="https://github.com/nikfot/netcdfella/issues">Request Feature</a>\n  </p>\n</div>\n\n<!-- TABLE OF CONTENTS -->\n<details>\n  <summary>Table of Contents</summary>\n  <ol>\n    <li>\n      <a href="#about-the-project">About The Project</a>\n      <ul>\n        <li><a href="#built-with">Built With</a></li>\n      </ul>\n    </li>\n    <li>\n      <a href="#getting-started">Getting Started</a>\n      <ul>\n        <li><a href="#prerequisites">Prerequisites</a></li>\n        <li><a href="#installation">Installation</a></li>\n      </ul>\n    </li>\n    <li><a href="#usage">Usage</a></li>\n    <li><a href="#roadmap">Roadmap</a></li>\n    <li><a href="#contributing">Contributing</a></li>\n    <li><a href="#license">License</a></li>\n    <li><a href="#contact">Contact</a></li>\n    <li><a href="#acknowledgments">Acknowledgments</a></li>\n  </ol>\n</details>\n\n<!-- ABOUT THE PROJECT -->\n\n## About The Project\n\nNetcdfella is a command line interface (cli) tool for converting netcdf files to ASCII and JPG/PNG format, as well as creating graphs based on the vectors of the file.\n\nNetcdfella supports two ways for converting documents at the time being:\n\n- single/multi convert\n- conversion upon creation\n\nDuring the run of the cli the output types, the mapping dimension and variable can be selected aling with other options.\n\n### single/multi convert\n\nYou can convert a single file or all netcdf files in a directory on your command. This is the case of the single/multi select.\n\n### conversion upon creation\n\nA directory can be selected to be watched by netcdfella. WHen a new netcdf file is created in that directory, it is automatically converted in the selected formats.\n\n#### qnotify\n\nIn order to handle any amount of input documents in the case of watch. A producer-consumer queue is created in the module qnotify using the inotify library. In this case the producers are the directory events. This creates a buffer that can handle the set amount of concurrent document creations.\n\nUse the `README.md` to get started.\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n### Built With\n\nThis section should list any major frameworks/libraries used to bootstrap your project. Leave any add-ons/plugins for the acknowledgements section. Here are a few examples.\n\n- [![Python][python]][python-url]\n- [![Poetry][poetry]][poetry-url]\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n<!-- GETTING STARTED -->\n\n## Getting Started\n\nThis is an example of how you may give instructions on setting up your project locally.\nTo get a local copy up and running follow these simple example steps.\n\n### Prerequisites\n\nThis software and tested for Python version 3.10.4.\n\nIf you need to manually install packages:\n\n- requirements\n  ```sh\n  pip install -r requirements.txt\n  ```\n\nIf you wish to make use of the pyproject.toml you need to install poetry:\n\n- poetry\n  ```sh\n  curl -sSL https://install.python-poetry.org | python3 -\n  ```\n\n### Installation\n\n_You can install from source or from pypi_\n\n#### Source\n\n1. Clone the repo\n   ```sh\n   git clone https://github.com/nikfot/netcdfella.git\n   ```\n2. Install using poetry\n   ```sh\n   poetry install\n   ```\n3. Run `netcdfella` to see the help with available choices:\n\n   ```sh\n   netcdfella\n   ```\n\n   or\n\n   ```sh\n   netcdfella --help\n   ```\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n<!-- USAGE EXAMPLES -->\n\n## Usage\n\nYou can use the netcdfella to get a better description of the available choices:\n\n    ```sh\n    netcdfella\n    ```\n\n    or\n\n    ```sh\n    netcdfella --help\n    ```\n\n### convert\n\nUse netcdfella to convert a single document or all documents in a directory:\n\n    ```sh\n    netcdfella convert "/path/to/documents" -k "ascii,graph,scatter,marks" -md "flashes" -mv "radiance"\n    ```\n\nAll choices for convert subcommand:\nOptions:\n-o, --output-dir TEXT set the output directory for converted files.\n-k, --output-kinds TEXT set the output kind for conversion.\n-md, --map-dimension TEXT set the dimenion to use for mapping.\n-mv, --map-variable TEXT set the variable to use for mapping.\n-e, --exclude-variables TEXT comma separated list of variables to be\nexcluded from conversion.\n--help Show this message and exit.\n\n### watch\n\nUse netcdfella to watch over a directory and convert files upon creation:\n\n    ```sh\n    netcdfella watch "/path/to/documents" -k "ascii,graph,scatter,marks" -md "flashes" -mv "radiance"\n    ```\n\nOptions:\n-o, --output-dir TEXT set the output directory for converted files.\n-k, --output-kinds TEXT set the output kind for conversion.\n-md, --map-dimension TEXT set the dimenion to use for mapping.\n-mv, --map-variable TEXT set the variable to use for mapping.\n-e, --exclude-variables TEXT comma separated list of variables to be\nexcluded from conversion.\n--help Show this message and exit.\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n<!-- ROADMAP -->\n\n## Roadmap\n\n- [x] Add cli conversion\n- [x] Add ascii,jpg/png, graphs\n- [ ] Add docker support\n- [ ] Check compatibility with more netcdf filetypes (current test METEOSAT 4th gen)\n- [ ] Add REST API Server\n- [ ] Add gui\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n<!-- CONTRIBUTING -->\n\n## Contributing\n\nThere are still a lot to be done, so if you find the project usefull please contribute you comments, ideas and code. It is **greatly appreciated**.\n\nIf you have a suggestion that would make this better, please fork the repo and create a pull request. You can also simply open an issue with the tag "enhancement".\nDon\'t forget to give the project a star! Thanks again!\n\n1. Fork the Project\n2. Create your Feature Branch (`git checkout -b feature/AmazingFeature`)\n3. Commit your Changes (`git commit -m \'Add some AmazingFeature\'`)\n4. Push to the Branch (`git push origin feature/AmazingFeature`)\n5. Open a Pull Request\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n<!-- LICENSE -->\n\n## License\n\nDistributed under the GPL License. See `LICENSE.md` for more information.\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n<!-- CONTACT -->\n\n## Contact\n\nNikos Fotiou - [@workaround18](https://twitter.com/workaround18) - nik_fot@hotmail.gr\n\nNetcdfella: [https://github.com/nikfot/netcdfella](https://github.com/nikfot/netcdfella)\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n<!-- ACKNOWLEDGMENTS -->\n\n## Acknowledgments\n\nThis tool is created as part of my postgrad studies "Space Technologies Applications and Services - STAR" programm at the National and Kapodestrian University of Athens Greece (NKUA). It was a semester project for the class "Space Image Processing" by professor Stavros Kolios.\n\n<p align="right">(<a href="#readme-top">back to top</a>)</p>\n\n[issues-shield]: https://img.shields.io/badge/github-issues-gray?style=for-the-badge&logo=Github\n[issues-url]: https://github.com/nikfot/netcdfella/issues\n[license-shield]: https://img.shields.io/pypi/l/netcdfella?style=plastic\n[license-url]: https://github.com/nikfot/netcdfella/blob/master/LICENSE.md\n[linkedin-shield]: https://img.shields.io/badge/-LinkedIn-black.svg?style=for-the-badge&logo=linkedin&colorB=555\n[linkedin-url]: https://www.linkedin.com/in/nikosfotiou/\n[netcdfella-logo]: assets/netcdfella_logo.png\n[python]: https://img.shields.io/badge/python-yellow?style=for-the-badge&logo=Python\n[python-url]: https://www.python.org/\n[poetry]: https://img.shields.io/badge/poetry-blue?style=for-the-badge&logo=Poetry\n[poetry-url]: https://upstreamsystems.atlassian.net/jira/your-work\n',
    'author': 'Nikos Fotiou',
    'author_email': 'nik_fot@hotmail.gr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://nikfot.github.io/netcdfella/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.10.4,<4.0.0',
}


setup(**setup_kwargs)



"""
maps module defines the available type of maps for netcdfella. 
Currently geostationary basemap is supported.
"""

import numpy as np
from mpl_toolkits.basemap import Basemap


class GeoStationaryMap(Basemap):
    def __init__(self, name, resolution="i", latitude0=0):
        self.name = name
        self.resolution = resolution
        self.latitude0 = latitude0
        self.satellite_height = None
        self.sphere = None

    def set_sat_height(self, height):
        "adjust the height of the satellite or leave blank for default"
        self.satellite_height = height
        return self

    def set_sphere(self, sphere):
        "set the sphere dimensions or leave blank for default"
        self.sphere = sphere
        return self

    def create_map(self):
        "create the basemap based on your previous settings"
        Basemap.__init__(
            self,
            projection="geos",
            resolution=self.resolution,
            rsphere=self.sphere,
            satellite_height=self.satellite_height,
            lat_0=self.latitude0,
            lon_0=0
        )
        return self

    def draw(self, kind=None):
        """
        choose the kind of map to draw.
        available: "basic","blue_marbel"
        """
        if kind == "blue_marble":
            self.bluemarble()
        else:
            self.draw_basic_earth()

    def draw_basic_earth(self, land="Linen", ocean="#AFDCEC"):
        "draw a basic style Earth globe"
        self.drawcoastlines(zorder=1)
        self.drawcountries(zorder=2)
        self.drawlsmask(
            land_color=land, ocean_color=ocean, zorder=0
        )
        self.drawmeridians(np.arange(0, 360, 30))
        self.drawparallels(np.arange(-90, 90, 30))
